/****************************************************************
Copyright (c) 1998 Sandro Sigala <ssigala@globalnet.it>.
All rights reserved.

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the name of the author not be used in
advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

The author disclaim all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall the author be liable for any
special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
****************************************************************/

#include "config.h"

#include <kapp.h>
#include <kmsgbox.h>
#include <qaccel.h>
#include <qbitmap.h>
#include <qfile.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qtstream.h>

#include "krepton.h"
#include "gamewidget.h"
#include "map.h"

GameWidget::GameWidget(QWidget *parent, const char *name)
        : QWidget(parent, name)
{
	in_game = false;
	in_pause = false;
	sprites = 0;
	maps = 0;

	QAccel *accel = new QAccel(this);
	accel->connectItem(accel->insertItem(Key_Up), this, SLOT(goUp()));
	accel->connectItem(accel->insertItem(Key_Down), this, SLOT(goDown()));
	accel->connectItem(accel->insertItem(Key_Left), this, SLOT(goLeft()));
	accel->connectItem(accel->insertItem(Key_Right), this, SLOT(goRight()));

	timerObjects = startTimer(200);
	timerMonsters = startTimer(300);
	timerEggs = startTimer(500);
	timerSecs = startTimer(1000);
	timerPlants = startTimer(3000);
}

void GameWidget::loadEpisode(const QString& path)
{
	loadSprites(path);
	loadMaps(path);
}

void GameWidget::loadSprites(const QString& path)
{
	QPixmap allsprites;

	allsprites.load(path + "/sprites.bmp");
	if (sprites) delete [] sprites;
	sprites = new QPixmap[num_sprites];

	for (int y = 0; y < 6; ++y)
		for (int x = 0; x < 8; ++x) {
			int i = y*8 + x; 
			sprites[i].resize(sprite_width, sprite_height);
			QPainter p;
			p.begin(&sprites[i]);
			p.drawPixmap(0, 0, allsprites,
				     x * sprite_width, y * sprite_height,
				     sprite_width, sprite_height);
			p.end();
		}
	sprites[Obj_Blip].setMask(sprites[Obj_Blip].createHeuristicMask());
	sprites[Obj_Blip2].setMask(sprites[Obj_Blip2].createHeuristicMask());
}

void GameWidget::loadMaps(const QString& path)
{
	QFile f(path + "/info");
	if (!f.open(IO_ReadOnly))
		fatal("Cannot open data files.\nHave you correctly installed KRepton?");
	QTextStream t(&f);
	QString name;
	name = t.readLine();
	t >> num_levels;
	f.close();

	if (maps) delete [] maps;
	maps = new Map[num_levels];

	QString p;
	for (int i = 0; i < num_levels; ++i) {
		p.sprintf("%s/map%d", (const char *)path, i + 1);
		maps[i].loadData(p);
		maps[i].do_sounds = &do_sounds;
	}
}

void GameWidget::startGame(const QString& episode, int level, int points)
{
	episode_name = episode;

	QString msg;
	msg.sprintf("Starting level %d of episode %s\n\nPassword is \"%s\"\n",
		    level + 1, (const char *)episode,
		    (const char *)maps[level].password);
	KMsgBox::message(this, "KRepton", msg);

	current_map = &maps[level];
	current_level = level;
	current_map->num_points = points;

	in_game = true;
	in_pause = false;
}

void GameWidget::startGamePassword(const QString& episode, const QString& password)
{
	for (int i = 0; i < num_levels; ++i)
		if (!stricmp(maps[i].password, password)) {
			startGame(episode, i);
			return;
		}
	KMsgBox::message(this, "KRepton", "Wrong password, sorry!");
}

void GameWidget::timerEvent(QTimerEvent *e)
{
	if (!in_game || in_pause)
		return;

	if (current_map->status_die || current_map->status_end_level) {
		in_game = false;
		if (current_map->status_die) {
			KMsgBox::message(this, "KRepton", "Died, sorry!");
			emit gameOver();
		} else {
			if (current_level + 1 < num_levels) {
				KMsgBox::message(this, "KRepton",
						 "Level finished!");
				startGame(episode_name, current_level + 1,
					  current_map->num_points);
			}
			else {
				KMsgBox::message(this, "KRepton",
						 "Episode finished!");
				emit gameOver();
			}
		}
		repaint(false);
		return;
	}

	int id = e->timerId();

	if (id == timerObjects) {
		current_map->updateObjects();
		repaint(false);
	} else if (id == timerMonsters) {
		current_map->updateMonsters();
		repaint(false);
	} else if (id == timerEggs) {
		current_map->updateEggs();
		current_map->updateObjects();
		repaint(false);
	} else if (id == timerSecs) {
		if (--current_map->num_secs <= 0) {
			in_game = false;
			KMsgBox::message(this, "KRepton", "Time out, sorry!");
			emit gameOver();
		}
	} else if (id == timerPlants) {
		current_map->updatePlants();
		repaint(false);
	}

	emit changedStats(current_map->num_diamonds, current_map->num_secs,
			  current_map->num_points, current_map->have_key);
}

void GameWidget::paintEvent(QPaintEvent *)
{
	if (!in_game) {
		QPixmap bg(kapp->kde_datadir() + "/krepton/data/bg.bmp");
		for (int y = 0; y < height() / bg.height() + 1; ++y)
			for (int x = 0; x < width() / bg.width() + 1; ++x)
				bitBlt(this, x*bg.width(), y*bg.height(), &bg);
		return;
	}

	QPixmap pm(size());
	QPainter p;
	p.begin(&pm);
	for (int y = 0; y < height() / sprite_height; y++)
		for (int x = 0; x < width() / sprite_width; x++) {
			SpriteObj obj;
			int mx = x + current_map->xpos - width() / sprite_width / 2;
			int my = y + current_map->ypos - height() / sprite_height / 2;
			switch (obj = current_map->xy(mx, my)) {
			case Obj_Broken_Egg1:
			case Obj_Broken_Egg2:
			case Obj_Broken_Egg3:
				obj = Obj_Broken_Egg;
				break;
			case Obj_Falling_Egg:
				obj = Obj_Egg;
				break;
			case Obj_Repton:
				obj = current_map->sprite;
				break;
			default:;	// Avoid warning
			}

			Monster *m;
			if ((m = current_map->findMonster(mx, my)) != 0
			    && m->type == Obj_Monster)
				obj = m->sprite;

			p.drawPixmap(x * sprite_width, y * sprite_height, sprites[obj]);

			if (m != 0 && m->type == Obj_Blip)
				p.drawPixmap(x * sprite_width, y * sprite_height, sprites[m->sprite]);
		}

	if (in_pause) {
		QPixmap pause(kapp->kde_datadir() + "/krepton/data/pause.bmp");
		p.drawPixmap((width()-pause.width())/2, 
			     (height()-pause.height())/2, pause);
	}

	p.end();
	bitBlt(this, 0, 0, &pm);
}

void GameWidget::goUp()
{
	if (!in_game || in_pause) return;
	current_map->goUp();
	repaint(false);
}

void GameWidget::goDown()
{
	if (!in_game || in_pause) return;
	current_map->goDown();
	repaint(false);
}

void GameWidget::goLeft()
{
	if (!in_game || in_pause) return;
	current_map->goLeft();
	repaint(false);
}

void GameWidget::goRight()
{
	if (!in_game || in_pause) return;
	current_map->goRight();
	repaint(false);
}
